package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.google.gson.GsonBuilder
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import com.mojang.blaze3d.platform.InputConstants
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import net.minecraft.ChatFormatting
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.packs.resources.ResourceManager
import net.minecraft.server.packs.resources.SimpleJsonResourceReloadListener
import net.minecraft.tags.TagKey
import net.minecraft.util.profiling.ProfilerFiller
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.ItemLike
import net.minecraftforge.event.AddReloadListenerEvent
import net.minecraftforge.event.entity.player.ItemTooltipEvent
import net.minecraftforge.registries.ForgeRegistries
import org.apache.logging.log4j.LogManager
import org.lwjgl.glfw.GLFW
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.SystemTime
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.drive.IMatteryDrive
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.formatMatterFull
import ru.dbotthepony.mc.otm.core.formatSiComponent
import ru.dbotthepony.mc.otm.core.isZero
import ru.dbotthepony.mc.otm.core.orNull
import ru.dbotthepony.mc.otm.core.registryName
import ru.dbotthepony.mc.otm.storage.ItemStackWrapper
import java.math.BigInteger
import java.util.Collections
import java.util.stream.Stream
import kotlin.math.pow

object MatterManager : SimpleJsonResourceReloadListener(GsonBuilder().setPrettyPrinting().disableHtmlEscaping().create(), "otm_matter") {
	const val DIRECTORY = "otm_matter"

	sealed class MutableEntry(
		val modificationChain: MutableList<ResourceLocation>,
		var matter: ImpreciseFraction,
		var complexity: Double,
		var priority: Int
	) {
		abstract fun freeze(): Entry
	}

	class MutableTagEntry(
		val tag: TagKey<Item>,
		modificationChain: MutableList<ResourceLocation>,
		matter: ImpreciseFraction,
		complexity: Double,
		priority: Int
	) : MutableEntry(modificationChain, matter, complexity, priority) {
		override fun freeze(): Entry {
			return TagEntry(tag, ImmutableList.copyOf(modificationChain), matter, complexity, priority)
		}
	}

	class MutableKeyEntry(
		val key: ResourceLocation,
		modificationChain: MutableList<ResourceLocation>,
		matter: ImpreciseFraction,
		complexity: Double,
		priority: Int
	) : MutableEntry(modificationChain, matter, complexity, priority) {
		override fun freeze(): Entry {
			return KeyEntry(key, ImmutableList.copyOf(modificationChain), matter, complexity, priority)
		}
	}

	sealed class Entry(
		val modificationChain: List<ResourceLocation>,
		final override val matter: ImpreciseFraction,
		final override val complexity: Double,
		val priority: Int,
	) : IMatterValue {
		companion object : Entry(listOf(), ImpreciseFraction.ZERO, 0.0, Int.MIN_VALUE)
	}

	class TagEntry(
		val tag: TagKey<Item>,
		modificationChain: List<ResourceLocation>,
		matter: ImpreciseFraction,
		complexity: Double,
		priority: Int,
	) : Entry(modificationChain, matter, complexity, priority) {
		val bound by lazy {
			val manager = ForgeRegistries.ITEMS.tags() ?: throw NullPointerException("Forge registry of Items has no tags!")
			manager.getTag(tag)
		}
	}

	class KeyEntry(
		val key: ResourceLocation,
		modificationChain: List<ResourceLocation>,
		matter: ImpreciseFraction,
		complexity: Double,
		priority: Int,
	) : Entry(modificationChain, matter, complexity, priority)

	var canCompute = false
		private set

	private val keyEntries = HashMap<ResourceLocation, KeyEntry>()
	private val tagEntries = ArrayList<TagEntry>()
	private val computedEntries = Reference2ObjectOpenHashMap<Item, Entry>()
	val computedEntriesView: Map<Item, Entry> = Collections.unmodifiableMap(computedEntries)

	/**
	 * Returns directly defined matter value
	 */
	@JvmStatic
	fun getDirect(value: Item): Entry {
		check(canCompute) { "MatteryManager is not ready to compute matter values" }

		val get = computedEntries.get(value)

		if (get != null) {
			return get
		}

		synchronized(computedEntries) {
			return computedEntries.computeIfAbsent(value, Reference2ObjectFunction {
				it as Item
				val key = it.registryName ?: throw NullPointerException("$it has no registry name!")

				val keyEntry = keyEntries[key]

				if (keyEntry != null) {
					return@Reference2ObjectFunction keyEntry
				}

				for (entry in tagEntries) {
					if (entry.bound.contains(it)) {
						return@Reference2ObjectFunction entry
					}
				}

				return@Reference2ObjectFunction Entry.Companion
			})
		}
	}

	@JvmStatic
	val computedEntriesSize: Int get() = computedEntries.size

	@JvmStatic
	fun stream(): Stream<Map.Entry<Item, Entry>> {
		return computedEntriesView.entries.stream().filter { it.value.hasMatterValue }
	}

	private var resolvedEverything = false

	/**
	 * Iterates through [ForgeRegistries.ITEMS] if it hadn't already and resolves everything
	 *
	 * THIS IS A VERY SLOW OPERATION TO PERFORM!
	 */
	@JvmStatic
	fun resolveEverything() {
		if (resolvedEverything) {
			return
		}

		val time = SystemTime()
		LOGGER.info("resolveEverything() was called")

		for (value in ForgeRegistries.ITEMS.values) {
			getValue(value)
		}

		resolvedEverything = true
		LOGGER.info("resolveEverything() finished in ${time.millis}ms")
	}

	/**
	 * Iterates through [ForgeRegistries.ITEMS] if it hadn't already and resolves everything,
	 * then returns stream of entries
	 *
	 * THIS IS A VERY SLOW OPERATION TO PERFORM!
	 */
	@JvmStatic
	fun resolveAndStream(): Stream<Map.Entry<Item, Entry>> {
		resolveEverything()
		return stream()
	}

	@JvmStatic
	fun getValue(value: Item): IMatterValue {
		val direct = getDirect(value)

		if (direct.hasMatterValue) {
			return direct
		}

		return RecipeResolverManager.getDetermined(value)
	}

	@JvmStatic
	fun hasMatterValue(value: Item): Boolean {
		return getValue(value).hasMatterValue
	}

	private fun getValue(value: ItemStack, level: Int): IMatterValue {
		if (value.isEmpty) {
			return IMatterValue.Companion
		}

		var matter: IMatterValue
		var undamagedMatterValue: IMatterValue

		if (value.item is IMatterItem)
			matter = (value.item as IMatterItem).getMatterValue(value) ?: IMatterValue.Companion
		else
			matter = getValue(value.item)

		if (matter.hasMatterValue && value.isDamageableItem) {
			val severity = 1.0 - value.damageValue.toDouble() / value.maxDamage.toDouble()
			undamagedMatterValue = matter
			matter = MatterValue(matter.matter * severity, matter.complexity * (0.5 + severity / 2))
		} else {
			undamagedMatterValue = matter
		}

		val matterCap = value.getCapability(MatteryCapability.MATTER).orNull()

		if (matterCap != null) {
			matter = MatterValue(matter.matter + matterCap.storedMatter, matter.complexity)
			undamagedMatterValue = MatterValue(undamagedMatterValue.matter + matterCap.storedMatter, undamagedMatterValue.complexity)
		}

		val drive = value.getCapability(MatteryCapability.DRIVE).orNull()

		if (drive != null && drive.storageType === OverdriveThatMatters.INSTANCE.ITEM_STORAGE()) {
			for (item in (drive as IMatteryDrive<ItemStackWrapper>).stacks) {
				val tuple = getValue(item.stack.stack, level + 1)

				if (tuple.hasMatterValue) {
					matter += tuple
					undamagedMatterValue += tuple
				}
			}
		}

		if (matter.hasMatterValue && value.count > 1) {
			return MatterValue(matter.matter + undamagedMatterValue.matter * (value.count - 1), matter.complexity + undamagedMatterValue.complexity * (value.count - 1))
		}

		return matter
	}

	/**
	 * Matter value for itemstack, accounting for capabilities and stack size
	 */
	@JvmStatic
	fun getValue(value: ItemStack): IMatterValue {
		return getValue(value, 0)
	}

	@JvmStatic
	fun hasMatterValue(value: ItemStack): Boolean {
		return getValue(value).hasMatterValue
	}

	@JvmStatic
	fun canDecompose(value: Item): Boolean {
		if (value is IMatterItem)
			return value.canDecompose(ItemStack.EMPTY)

		return getValue(value).hasMatterValue
	}

	@JvmStatic
	fun canDecompose(value: ItemStack): Boolean {
		var can = true

		if (value.item is IMatterItem)
			can = (value.item as IMatterItem).canDecompose(value)

		can = can && (value.getCapability(MatteryCapability.MATTER).orNull()?.storedMatter ?: ImpreciseFraction.ZERO).isZero
		can = can && (value.getCapability(MatteryCapability.DRIVE).orNull()?.storedCount ?: BigInteger.ZERO).isZero

		return can && getValue(value).hasMatterValue
	}

	@JvmStatic
	fun getResearchAdvance(complexity: Double): Double {
		if (complexity <= 0.0) {
			return 0.0
		}

		return (1.0 / (complexity.pow(1.0 / 3.0) * 12.0).coerceAtLeast(0.01)).coerceAtMost(0.2).coerceAtLeast(0.025)
	}

	@JvmStatic
	fun getResearchAdvance(stack: ItemStack): Double {
		return getResearchAdvance(getValue(stack).complexity)
	}

	@JvmStatic
	fun getResearchAdvance(stack: ItemLike): Double {
		return getResearchAdvance(getValue(stack.asItem()).complexity)
	}

	fun tooltipEvent(event: ItemTooltipEvent) {
		val window = minecraft.window.window

		if (InputConstants.isKeyDown(window, GLFW.GLFW_KEY_LEFT_SHIFT) || InputConstants.isKeyDown(window, GLFW.GLFW_KEY_RIGHT_SHIFT)) {
			val matter = getValue(event.itemStack)

			if (matter.hasMatterValue) {
				if (matter.complexity >= 1.0E-3) {
					event.toolTip.add(
						TranslatableComponent("otm.gui.matter.format_and_complexity",
							matter.matter.formatMatterFull(),
							matter.complexity.formatSiComponent(TranslatableComponent("otm.gui.ticks"))
						).withStyle(ChatFormatting.AQUA))
				} else {
					event.toolTip.add(matter.matter.formatMatterFull().withStyle(ChatFormatting.AQUA))
				}
			}

			if (event.flags.isAdvanced) {
				val commentary = RecipeResolverManager.getCommentary(event.itemStack.item)?.withStyle(ChatFormatting.DARK_GRAY)

				if (commentary != null) {
					event.toolTip.add(commentary)
				}
			}
		}
	}

	fun reloadEvent(event: AddReloadListenerEvent) {
		event.addListener(this)
	}

	private val LOGGER = LogManager.getLogger()

	override fun prepare(
		resourceManager: ResourceManager,
		profilerFiller: ProfilerFiller
	): MutableMap<ResourceLocation, JsonElement> {
		val time = System.nanoTime()
		LOGGER.info("Beginning preparations for loading matter values")
		profilerFiller.push("otm_matter_registry_prepare")

		try {
			val prep = super.prepare(resourceManager, profilerFiller)
			val diff = System.nanoTime() - time
			LOGGER.info("Finished preparations for loading matter values, took ${diff / 1_000_000L}ms")
			return prep
		} finally {
			profilerFiller.pop()
		}
	}

	override fun apply(
		map: MutableMap<ResourceLocation, JsonElement>,
		resourceManager: ResourceManager,
		profilerFiller: ProfilerFiller
	) {
		val time = System.nanoTime()
		LOGGER.info("Beginning loading matter values")
		profilerFiller.push("otm_matter_registry")

		try {
			canCompute = false
			resolvedEverything = false

			keyEntries.clear()
			tagEntries.clear()
			computedEntries.clear()

			val add = ArrayList<Pair<InsertAction, ResourceLocation>>()
			val addOrReplace = ArrayList<Pair<InsertAction, ResourceLocation>>()
			val update = ArrayList<Pair<UpdateAction, ResourceLocation>>()
			val delete = ArrayList<Pair<DeleteAction, ResourceLocation>>()

			for ((key, json) in map) {
				if (json !is JsonObject) {
					throw JsonParseException("Matter value $key has invalid type: ${json::class.qualifiedName} ($json)")
				}

				when (val action = json["action"]?.asString?.lowercase() ?: throw JsonParseException("Missing `action` json element in $key. Possible values are: (INSERT, DELETE and UPDATE (case-insensitive))")) {
					"insert" -> {
						val value = InsertAction(json)

						if (value.replaceIfExists) {
							addOrReplace.add(value to key)
						} else {
							add.add(value to key)
						}
					}

					"update" -> update.add(UpdateAction(json) to key)
					"delete" -> delete.add(DeleteAction(json) to key)
					else -> throw JsonParseException("Unknown action $action of $key")
				}
			}

			val registry = ArrayList<MutableEntry>()

			for (action in add)
				action.first.update(registry, action.second)

			for (action in addOrReplace)
				action.first.update(registry, action.second)

			for (action in delete)
				action.first.update(registry, action.second)

			for (action in update)
				action.first.update(registry, action.second)

			val tags = HashMap<TagKey<Item>, MutableTagEntry>()
			val keys = HashMap<ResourceLocation, MutableKeyEntry>()

			for (entry in registry) {
				if (entry is MutableTagEntry) {
					val existing = tags[entry.tag]

					if (existing == null || existing.priority < entry.priority) {
						tags[entry.tag] = entry
					}
				} else if (entry is MutableKeyEntry) {
					val existing = keys[entry.key]

					if (existing == null || existing.priority < entry.priority) {
						keys[entry.key] = entry
					}
				}
			}

			for (key in keys.values) {
				keyEntries[key.key] = key.freeze() as KeyEntry
			}

			for (tag in tags.values) {
				tagEntries.add(tag.freeze() as TagEntry)
			}

			tagEntries.sortBy { it.priority }

			canCompute = true
		} finally {
			profilerFiller.pop()
		}

		val diff = System.nanoTime() - time
		LOGGER.info("Finished loading matter values, took ${diff / 1_000_000L}ms")
	}
}
