package ru.dbotthepony.mc.otm.matter

import com.google.common.collect.ImmutableList
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.JsonParseException
import net.minecraft.resources.ResourceLocation
import net.minecraft.tags.TagKey
import net.minecraft.world.item.Item
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.integerDivisionDown
import ru.dbotthepony.mc.otm.core.integerDivisionUp
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.data.stream

class UpdateAction : AbstractRegistryAction {
	enum class Function {
		ADD {
			override fun updateValue(self: Int, other: Int): Int = 												self + other
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self + other
			override fun updateValue(self: Double, other: Double): Double = 									self + other
		},
		SUBTRACT {
			override fun updateValue(self: Int, other: Int): Int = 												self - other
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self - other
			override fun updateValue(self: Double, other: Double): Double = 									self - other
		},
		MULTIPLY {
			override fun updateValue(self: Int, other: Int): Int = 												self * other
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self * other
			override fun updateValue(self: Double, other: Double): Double = 									self * other
		},
		DIVIDE {
			override fun updateValue(self: Int, other: Int): Int = 												self / other
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self / other
			override fun updateValue(self: Double, other: Double): Double = 									self / other
		},
		DIVIDE_UP {
			override fun updateValue(self: Int, other: Int): Int = integerDivisionUp(self, other)
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	throw JsonParseException(
				"Integer division up is available only for integers"
			)
			override fun updateValue(self: Double, other: Double): Double = 									throw JsonParseException(
				"Integer division up is available only for integers"
			)
		},
		DIVIDE_DOWN {
			override fun updateValue(self: Int, other: Int): Int = integerDivisionDown(self, other)
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	throw JsonParseException(
				"Integer division down is available only for integers"
			)
			override fun updateValue(self: Double, other: Double): Double = 									throw JsonParseException(
				"Integer division down is available only for integers"
			)
		},
		MODULO {
			override fun updateValue(self: Int, other: Int): Int = 												self % other
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self % other
			override fun updateValue(self: Double, other: Double): Double = 									self % other
		},
		AT_LEAST {
			override fun updateValue(self: Int, other: Int): Int = 												self.coerceAtLeast(other)
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self.coerceAtLeast(other)
			override fun updateValue(self: Double, other: Double): Double = 									self.coerceAtLeast(other)
		},
		AT_MOST {
			override fun updateValue(self: Int, other: Int): Int = 												self.coerceAtMost(other)
			override fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction = 	self.coerceAtMost(other)
			override fun updateValue(self: Double, other: Double): Double = 									self.coerceAtMost(other)
		},
		;

		protected abstract fun updateValue(self: Int, other: Int): Int
		protected abstract fun updateValue(self: ImpreciseFraction, other: ImpreciseFraction): ImpreciseFraction
		protected abstract fun updateValue(self: Double, other: Double): Double

		fun <T : Number> updateValue(self: T, other: T): T {
			return when (self) {
				is ImpreciseFraction -> updateValue(self, other)
				is Double -> updateValue(self, other)
				is Int -> updateValue(self, other)
				else -> throw RuntimeException("Unknown number type: ${self::class.qualifiedName}")
			}
		}
	}

	class BoundFunction<T : Number>(val function: Function, val value: T) {
		fun apply(self: T): T {
			return function.updateValue(self, value)
		}

		fun toJson(): JsonObject {
			return JsonObject().also {
				it["type"] = function.name

				if (value is Int || value is Double)
					it["value"] = value
				else
					it["value"] = value.toString()
			}
		}
	}

	companion object {
		inline fun <reified T : Number> getValue(value: JsonElement): T {
			return when (T::class) {
				ImpreciseFraction::class -> ImpreciseFraction(value.asString) as T
				Double::class -> value.asDouble as T
				Int::class -> value.asInt as T
				else -> throw RuntimeException("Unknown number type: ${T::class.qualifiedName}")
			}
		}

		inline fun <reified T : Number> deserializeFunctionTree(input: JsonElement?, name: String): List<BoundFunction<T>> {
			if (input == null) {
				return listOf()
			}

			if (input !is JsonArray) {
				throw JsonParseException("Expected $name to be JsonArray, ${input::class.qualifiedName} given")
			}

			return input.stream().map {
				if (it !is JsonObject) {
					throw JsonParseException("All elements in function tree must be JsonObjects, ${it::class.qualifiedName} given")
				}

				val type = it["type"]?.asString ?: throw JsonParseException("Invalid `type` value in function object")
				val value = it["value"] ?: throw JsonParseException("Missing `value` value in function object")

				val parsedValue: T = getValue(value)
				val fn: Function

				try {
					fn = Function.valueOf(type.uppercase())
				} catch (err: NoSuchElementException) {
					throw JsonParseException("No such function: $type", err)
				}

				return@map BoundFunction(fn, parsedValue)
			}.collect(ImmutableList.toImmutableList())
		}

		fun <T : Number> apply(value: T, funcs: Collection<BoundFunction<T>>): T {
			if (funcs.isEmpty()) {
				return value
			}

			var newValue = value

			for (func in funcs) {
				newValue = func.apply(newValue)
			}

			return newValue
		}
	}

	val matterFunctions: List<BoundFunction<ImpreciseFraction>>
	val complexityFunctions: List<BoundFunction<Double>>
	val priorityFunctions: List<BoundFunction<Int>>

	constructor(json: JsonObject) : super(json) {
		matterFunctions = deserializeFunctionTree(json["matter_functions"], "matter_functions")
		complexityFunctions = deserializeFunctionTree(json["complexity_functions"], "complexity_functions")
		priorityFunctions = deserializeFunctionTree(json["priority_functions"], "priority_functions")
	}

	constructor(
		tag: TagKey<Item>,
		matter: ImpreciseFraction? = null,
		complexity: Double? = null,
		priority: Int? = null,
		errorOnFailure: Boolean = false,
		matterFunctions: List<BoundFunction<ImpreciseFraction>> = ImmutableList.of(),
		complexityFunctions: List<BoundFunction<Double>> = ImmutableList.of(),
		priorityFunctions: List<BoundFunction<Int>> = ImmutableList.of(),
	) : super(tag = tag, matter = matter, complexity = complexity, priority = priority, errorOnFailure = errorOnFailure) {
		check(matter != null || complexity != null || priority != null || matterFunctions.isNotEmpty() || complexityFunctions.isNotEmpty() || priorityFunctions.isNotEmpty()) {
			"Can't update existing entries without anything specified"
		}

		this.matterFunctions = ImmutableList.copyOf(matterFunctions)
		this.complexityFunctions = ImmutableList.copyOf(complexityFunctions)
		this.priorityFunctions = ImmutableList.copyOf(priorityFunctions)
	}

	constructor(
		key: ResourceLocation,
		matter: ImpreciseFraction? = null,
		complexity: Double? = null,
		priority: Int? = null,
		errorOnFailure: Boolean = false,
		matterFunctions: List<BoundFunction<ImpreciseFraction>> = ImmutableList.of(),
		complexityFunctions: List<BoundFunction<Double>> = ImmutableList.of(),
		priorityFunctions: List<BoundFunction<Int>> = ImmutableList.of(),
	) : super(key = key, matter = matter, complexity = complexity, priority = priority, errorOnFailure = errorOnFailure) {
		check(matter != null || complexity != null || priority != null || matterFunctions.isNotEmpty() || complexityFunctions.isNotEmpty() || priorityFunctions.isNotEmpty()) {
			"Can't update existing entries without anything specified"
		}

		this.matterFunctions = ImmutableList.copyOf(matterFunctions)
		this.complexityFunctions = ImmutableList.copyOf(complexityFunctions)
		this.priorityFunctions = ImmutableList.copyOf(priorityFunctions)
	}

	override fun toJson(): JsonObject {
		return super.toJson().also {
			it["action"] = "update"

			if (matterFunctions.isNotEmpty()) {
				it["matter_functions"] = JsonArray().also {
					for (fn in matterFunctions) {
						it.add(fn.toJson())
					}
				}
			}

			if (complexityFunctions.isNotEmpty()) {
				it["complexity_functions"] = JsonArray().also {
					for (fn in complexityFunctions) {
						it.add(fn.toJson())
					}
				}
			}

			if (priorityFunctions.isNotEmpty()) {
				it["priority_functions"] = JsonArray().also {
					for (fn in priorityFunctions) {
						it.add(fn.toJson())
					}
				}
			}
		}
	}

	fun apply(value: MatterManager.MutableEntry, modifier: ResourceLocation) {
		if (matterFunctions.isNotEmpty())
			value.matter = apply(value.matter, matterFunctions)
		else if (matter != null)
			value.matter = matter

		if (complexityFunctions.isNotEmpty())
			value.complexity = apply(value.complexity, complexityFunctions)
		else if (complexity != null)
			value.complexity = complexity

		if (priorityFunctions.isNotEmpty())
			value.priority = apply(value.priority, priorityFunctions)
		else if (priority != null)
			value.priority = priority

		value.modificationChain.add(modifier)
	}

	override fun update(registry: MutableCollection<MatterManager.MutableEntry>, modifier: ResourceLocation) {
		if(!(
			matter != null ||
			complexity != null ||
			priority != null ||
			matterFunctions.isNotEmpty() ||
			complexityFunctions.isNotEmpty() ||
			priorityFunctions.isNotEmpty()
		)) {
			throw JsonParseException("Can't update existing entries without anything specified (for $modifier)")
		}

		if (matter != null && matterFunctions.isNotEmpty()) {
			throw JsonParseException("Can't have both matter and matter functions specified at the same time (for $modifier)")
		}

		if (complexity != null && complexityFunctions.isNotEmpty()) {
			throw JsonParseException("Can't have both complexity and complexity functions specified at the same time (for $modifier)")
		}

		if (priority != null && priorityFunctions.isNotEmpty()) {
			throw JsonParseException("Can't have both priority and priority functions specified at the same time (for $modifier)")
		}

		if (!checkConditions()) {
			return
		}

		if (key != null) {
			val iterator = registry.iterator()

			for (value in iterator) {
				if (value is MatterManager.MutableKeyEntry && value.key == key) {
					apply(value, modifier)
					return
				}
			}

			fail { "Could not find matter value with key $key" }
		} else {
			val iterator = registry.iterator()

			for (value in iterator) {
				if (value is MatterManager.MutableTagEntry && value.tag == tag) {
					apply(value, modifier)
					return
				}
			}

			fail { "Could not find matter value with tag $tag" }
		}
	}
}
